# frozen_string_literal: true

require 'English'
require 'create_github_release/task_base'

module CreateGithubRelease
  module Tasks
    # Update the gem version using Bump
    #
    # @api public
    #
    class UpdateVersion < TaskBase
      # Update the gem version using Bump
      #
      # @example
      #   require 'create_github_release'
      #
      #   options = CreateGithubRelease::CommandLineOptions.new { |o| o.release_type = 'major' }
      #   project = CreateGithubRelease::Project.new(options)
      #   task = CreateGithubRelease::Tasks::UpdateVersion.new(project)
      #   begin
      #     task.run
      #     puts 'Task completed successfully'
      #   rescue SystemExit
      #     puts 'Task failed'
      #   end
      #
      # @return [void]
      #
      # @raise [SystemExit] if the task fails
      #
      def run
        return if project.first_release?

        print 'Updating version...'
        bump_version
        stage_version_file
      end

      private

      # Update the version using bump
      # @return [void]
      # @api private
      def bump_version
        `bump #{project.release_type} --no-commit`
        error 'Could not bump version' unless $CHILD_STATUS.success?
      end

      # Return the path the the version file using bump
      # @return [String]
      # @api private
      def bump_version_file
        output = `bump file`
        error 'Bump could determine the version file' unless $CHILD_STATUS.success?

        output.lines.last.chomp
      end

      # Identify the version file using bump and stage the change to it
      # @return [void]
      # @api private
      def stage_version_file
        version_file = bump_version_file
        `git add "#{version_file}"`
        if $CHILD_STATUS.success?
          puts 'OK'
        else
          error "Could not stage changes to #{version_file}"
        end
      end
    end
  end
end
